//	FindinSite-CD-JS search code
//	Copyright (C) 2016 PHD Computer Consultants Ltd
//
//	Alter current version below
//		7.014	5-Jul-2016
//
//	The search database JS files can be loaded in two ways:
//	A:	Specify a list of files in a global array named findinsitedbs, eg specified in a JS file like this
//				<script src="search_dbs.js"></script>
//			whose contents looks like this:
//				var findinsitedbs = [
//					{ src:"db1.js", id:"varname"},
//				]
//			The script file paths have findinsite.config.baseurl prepended - this allows the script to be loaded correctly from different directories
//	B:	Use one or more script tags in the calling page like this:
//				<script async class="findinsitedb" src="db1.js" id="dbvarname" data-baseurl="" data-weight=""></script>
//			The script must be after this file is included - and async is optional but recommended
//			If you need to specify a path to prepend to the search results, give id and data-baseurl.
//
//	Usage:	var results = findinsite.search(params);
//													where params.search = 'search words'
//																params.showProgress = optional function to show progress
//																params.searchDone = required function to show results
//													and findinsite.databases[n].selected determines if a subset is searched
//					searchDone is called with results as follows:
//					{	error:			false or string
//						expression:	array of SearchWord-s { wordfound:true|false, given:string, w:array of synonyms etc }
//						hits:				array of [pageno,db,relevance] - subtact 1 from pageno to get index into db.pages
//						narrative:	null or array of strings
//
//	Currently requires jQuery to handle events triggered by database load(s)
//	Alternative approach is outlined here: http://blog.garstasio.com/you-dont-need-jquery/events/

////////////////////////////////////////////
// Missing:								stop words
//												wild cards
//												222 in wizard
//												??? here
// Deliberately missing:	different cases
// Suggestions:						suggested words
// Added:									cope if word not found

/*global window,jQuery */
/*global fisLogging,fisFreezeDb */
/*jshint bitwise: false*/
/*jslint continue: true, plusplus: true, todo: true, vars: true, white: true */

/** @namespace */
var findinsite = function ($) {
	"use strict";

	// PRIVATE variables made PUBLIC at end

	var version = "v7.014";
	var versionDate = new Date(2016, 7 - 1, 5);	// Month in range 0..11

	var databases = [];	// dbs actually loaded, ie not findinsitedbs

	var config = {
		rules: false,
		baseurl: '',
		dbcount: 0
	};

	// PRIVATE variables and functions - some made public at end

	var isUnicodeNumber, isUnicodeLetter, isUnicodeSeparator, isUnicodeCharLatin, toUnicodeLowerCase, doSearch, stringHashCode;

	var deferSearch = false;

	// MODULE WORKING VARIABLES: assumes we are not called re-entrantly
	var thisdb;				// current search database
	var showProgress;	// callback to show progress
	var searchDone;		// report results

	// Public variables/constants
	var OP_AND = 0x00;	// Acts as if AND isn't there
	var OP_NONE = OP_AND;
	var OP_OR = 0x01;
	var OP_NOT = 0x10;
	var OP_CONTIG = 0x20;

	var lcAND = "and";
	var lcOR = "or";
	var lcNOT = "not";

	/////////////////////////////////////
	/**
	 * Show message in console if global fisLogging defined.
	 *
	* @param {string} msg - Message to show
	* @private
	*/
	function log(msg) {
		if (typeof fisLogging !== 'undefined') {
			if (fisLogging && window.console && window.console.log) {
				window.console.log(new Date().toISOString() + " " + msg);
			}
		}
	}

	/////////////////////////////////////
	function showLoadingProgress() {
		if (!showProgress) { return; }
		var loadedDbcount = databases.length;
		if (loadedDbcount == config.dbcount) {
			showProgress("Searching");
		} else {
			showProgress("Loaded " + loadedDbcount + " of " + config.dbcount + " search databases");
		}
	}

	/////////////////////////////////////

	function getGlobalName(ofind) {
		try {
			for (var o in window) {
				if (window[o] === ofind) {
					return o;
				}
			}
		} catch (e) { }
		return null;
	}

	/////////////////////////////////////
	// Startup: initiate loading of databases if necessary
	$(function () {
		if ((typeof findinsitedbs !== 'undefined') && $.isArray(findinsitedbs)) {
			// findinsitedbs defines which database JS files to load
			findinsite.config.dbcount = findinsitedbs.length;

			for (var dbno = 0; dbno < findinsitedbs.length; dbno++) {
				// http://stackoverflow.com/a/25897833/1541015
				var dbdef = findinsitedbs[dbno];
				var s = document.createElement('script');
				s.type = "text/javascript";
				s.async = true;
				dbdef.src = findinsite.config.baseurl + dbdef.src;
				s.src = dbdef.src;
				s.id = dbdef.id;
				var baseurl = '';
				var endbaseurl = dbdef.src.lastIndexOf("/");
				if (endbaseurl !== -1) {
					baseurl = dbdef.src.substr(0, endbaseurl + 1);
				}
				s.setAttribute("data-baseurl", baseurl);
				s.setAttribute("data-weight", dbno);
				var fs = document.getElementsByTagName('script')[0];
				fs.parentNode.insertBefore(s, fs);
			}

		} else {
			// database JS files loaded by calling page
			findinsite.config.dbcount = $('.findinsitedb').length;
			if (findinsite.config.dbcount == 0) {

				if (window.console && window.console.log) {
					window.console.log("findinsite: no search database JS files defined.");
				}
				findinsite.config.dbcount = 1;
			}
			// All dbs may already have loaded, so:
			processIfAllDbsLoaded();
		}
	}
	);

	/////////////////////////////////////
	// Handle database loaded event
	// If search pending: when all databases loaded, do search
	$(document).on("findinsiteDbLoaded", function (e) {
		var db = e.db;
		db.selected = true;
		var dbno = databases.push(db) - 1;

		db.baseurl = $('#' + db.varname).data('baseurl');
		if (typeof db.baseurl === 'undefined') { db.baseurl = ""; }

		db.weight = $('#' + db.varname).data('weight');

		//console.log("fiscdjs: " + new Date().toISOString() + " db loaded " + db.varname + " - " + db.description + " - " + db.baseurl);
		showLoadingProgress();
		processIfAllDbsLoaded();
	});

	/////////////////////////////////////
	// If all expected databases loaded, trigger event and do any deferred search
	function processIfAllDbsLoaded() {
		if (databases.length == config.dbcount) {
			databases.sort(function (dba, dbb) { return dba.weight - dbb.weight });
			$.event.trigger({ type: "findinsiteDbAllLoaded" });
			if (deferSearch) {
				var params = deferSearch;
				deferSearch = false;
				search(params);
			}
		}
	}

	/////////////////////////////////////
	/*function arrayUnique(array) {
	var a = array.concat();
	for (var i = 0; i < a.length; ++i) {
	for (var j = i + 1; j < a.length; ++j) {
	if (a[i] === a[j]) {
	a.splice(j--, 1);
	}
	}
	}
	return a;
	}*/

	/////////////////////////////////////
	// Remove duplicate pagenos from given array
	//	each element either	[pageno,wordposn1...]
	//		From:	[[4, 370], [15, 814, 1779], [29, 1528, 1531, 2272], [34, 447, 3119, 3201], [42, 908], [15, 1715, 1732], [42, 799]]
	//		To:		[[4, 370], [15, 814, 1779, 1715, 1732], [29, 1528, 1531, 2272], [34, 447, 3119, 3201], [42, 908, 799]]
	//	or [pageno,wordcount]
	//		From:	[[4, 1], [15, 2], [29, 3], [34, 3], [42, 1], [15, 2], [42, 1]]
	//		To:		[[4, 1], [15, 4], [29, 3], [34, 3], [42, 2]]
	function plNormalise(arr) {
		var a, i, j, k;
		a = arr.concat();
		for (i = 0; i < a.length; i++) {
			for (j = i + 1; j < a.length; j++) {
				if (a[i][0] === a[j][0]) {
					if (thisdb.haswordposns) {	// merge word positions
						for (k = 1; k < a[j].length; k++) {
							a[i].push(a[j][k]);
						}
					} else {	// add word counts
						a[i][1] += a[j][1];
					}
					a.splice(j--, 1);
				}
			}
		}
		return a;
	}
	/////////////////////////////////////
	// Remove duplicate pagenos from given combined array
	//	each element either	[pageno,prio,wordposn1...]
	//		From:	[[4, 2, 370], [15, 3, 814, 1779], [29, 3, 1528, 1531, 2272], [34, 2, 447, 3119, 3201], [42, 2, 908], [15, 2, 1715, 1732], [42, 2, 799]]
	//		To:		[[4, 2, 370], [15, 2, 814, 1779, 1715, 1732], [29, 3, 1528, 1531, 2272], [34, 2, 447, 3119, 3201], [42, 2, 908, 799]]
	//	or [pageno,prio,wordcount]
	//		From:	[[4, 2, 1], [15, 3, 2], [29, 3, 3], [34, 2, 3], [42, 2, 1], [15, 2, 2], [42, 2, 1]]
	//		To:		[[4, 2, 1], [15, 2, 4], [29, 3, 3], [34, 2, 3], [42, 2, 2]]
	function plNormaliseAndSortPositions(arr) {
		var a, i, j, k, pp, pageno, prio;
		a = arr.concat();
		for (i = 0; i < a.length; i++) {
			for (j = i + 1; j < a.length; j++) {
				if (a[i][0] === a[j][0]) {
					a[i][1] = Math.min(a[i][1], a[j][1]);
					if (thisdb.haswordposns) {	// merge word positions
						for (k = 2; k < a[j].length; k++) {
							a[i].push(a[j][k]);
							pp = a[i].splice(0, 2);
							pageno = pp[0];
							prio = pp[1];
							a[i].sort(function (a, b) { return a - b; });
							a[i].unshift(pageno, prio);
						}
					} else {	// add word counts
						a[i][2] += a[j][2];
					}
					a.splice(j--, 1);
				}
			}
		}
		return a;
	}

	/////////////////////////////////////
	function deepArrayCopy(arr) {
		var rarr, ix;
		if (Object.prototype.toString.call(arr) !== "[object Array]") { return arr; }
		rarr = [];
		for (ix = 0; ix < arr.length; ix++) {
			rarr.push(deepArrayCopy(arr[ix]));
		}
		return rarr;
	}

	/////////////////////////////////////
	function deepFreezeArray(arr) {
		var ix;
		//log(typeof arr + " - " +arr);
		if (typeof arr !== "object") { return; }
		Object.freeze(arr);
		if (Object.prototype.toString.call(arr) !== "[object Array]") { return; }

		for (ix = 0; ix < arr.length; ix++) {
			deepFreezeArray(arr[ix]);
		}
	}

	/////////////////////////////////////
	var SearchWord = function (word, lcword, op, useRules) {
		if (typeof useRules === 'undefined') { useRules = true; }
		//log("SearchWord: "+word+ " op "+op+" rules "+useRules);
		this.given = word;
		this.op = op;
		this.wds = [lcword];
		this.wdsfound = [];
		this.anyfound = false;

		if (useRules) {
			if (config.rules) {
				config.rules.generate(lcword, this.wds);
			}
		}
		//log(this.w);
	};

	/////////////////////////////////////
	function storeWord(narrative, expression, word, op, incontig, useRules) {
		var lcword = toUnicodeLowerCase(word);
		if (!incontig) {
			if (lcword == lcAND) {
				return op;
			}
			if (lcword == lcOR) {
				op = OP_OR;
				return op;
			}
			if (lcword == lcNOT) {
				op |= OP_NOT;
				return op;
			}
		}
		if ((expression.length === 0) && ((op & OP_OR) === OP_OR)) {
			narrative.push("Initial OR ignored");
			op &= ~OP_OR;
		}
		if ((op & OP_NOT) && incontig) {
			throw "NOT not permited before \"contiguous words\"";
		}

		if (incontig) { op |= OP_CONTIG; }
		expression.push(new SearchWord(word, lcword, op, useRules));
		return OP_NONE;
	}

	/////////////////////////////////////
	/**
   * Start search; call searchDone when done
   *
   * @param {object}	parameters
	 *		search				{text}			- Text to search for
	 *		showProgress	{function}	- Callback to show search progress
	 *		searchDone		{function}	- Callback to show results
   */

	function search(params) {
		if (typeof params.searchDone !== 'function') { return; }
		searchDone = params.searchDone;

		showProgress = false;
		if (typeof params.showProgress === 'function') { showProgress = params.showProgress; }

		try {
			doSearch(params);
		}
		catch (err) {
			searchDone({
				error: err
			});
		}
	}

	/////////////////////////////////////
	//	Search all available databases, one at a time then combine/sort results
	doSearch = function (params) {

		var given = params.search;
		deferSearch = false;

		if (databases.length > config.dbcount) {
			throw "Too many databases loaded: " + databases.length + ". Expected: " + config.dbcount;
		} else if (databases.length != config.dbcount) {
			deferSearch = params;
			showLoadingProgress();
			return;
		}

		if (!given) { given = ''; }
		//log("Search=" + given);
		given = given.trim();
		if (given.length === 0) { throw 'No search'; }

		var results = {
			error: false,
			hits: [],
			narrative: []
		};

		var isAllSearch = false;
		if (search === '*') {
			isAllSearch = true;
			results.expression = [{ given: given, wordfound: true }];
		}

		// Search each database, adding hits to results.hits
		var dbcount = databases.length;
		var anydbselected = false;
		for (var dbno = 0; dbno < dbcount; dbno++) {

			thisdb = databases[dbno];
			// 7.014 always search all databases, and then filter: if (!thisdb.selected) continue;
			anydbselected = true;
			searchOneDatabase(given, results, dbno);

		}
		if (!anydbselected) {
			results.error = "No subset selected";
		}

		// Sort to have hits with highest relevance first
		if (!isAllSearch) {
			results.hits.sort(function (a, b) {
				return b[2] - a[2];
			});
		}
		searchDone(results);
	};

	/////////////////////////////////////
	// Each database has a list of pages and a words hash table, with each entry like this:
	//	["word",[0,[[1,1214]]],[3,[[1,60,116],[11,60,116]]]],
	// where the word is followed by one or more priority-page-lists (ppl)
	// where	the priority is int 0 to 3
	//				followed by one or more page-lists
	// where page-list is either	[pageno,word position list...]		if thisdb.haswordposns	true
	//										or			[pageno,wordcount]																				false
	//
	// Search process is:
	//
	//	- If search is * then return all pages in database immediately
	//
	//	- Parse input into a series of SearchWord-s
	//		where searchWord.given is eg test 
	//		and		searchWord.wdsfound has alt-words that match by rule eg test tests tested testing unless given in single-quote eg 'test'
	//
	//	- For each SearchWord, for each alt-word, if alt-word in database, store ppl-s in searchWord.pl
	//
	//	- For each SearchWord, merge priority-page-lists to end up with searchWord.combined
	//			combined has array of	[pageno,prio,wordposn1...]
	//												or	[pageno,prio,totalwordcount]
	//
	//	- Merge the SearchWord.combined

	function searchOneDatabase(given, results, dbno) {
		var doFreeze, hashtableentries, allpageix, expr, word, chno, nch, useRules, eno, searchWord;
		var wno, lcword, hashcode, wordList, prioList, pr0no, plno, pln, prno, prlist, prio, foundprio, pr0list;
		var hitinfo, anyhits, tomerge, pageix, mergeix, pageno, mergepageno, merged, incontig;
		var i, pl, bestprio, totalwordcount;

		///////////////
		// Optionally freeze the db (to ensure that nothing is changed that will affect future searches)
		if ((typeof fisFreezeDb !== 'undefined') && fisFreezeDb) {
			doFreeze = true;
			if (Object.isFrozen && Object.isFrozen(thisdb)) {
				doFreeze = false;
			}
			if (doFreeze && Object.freeze) {
				log("FREEZE DB");
				Object.freeze(thisdb);
				deepFreezeArray(thisdb.pages);
				deepFreezeArray(thisdb.words);
			}
		}

		hashtableentries = thisdb.words.length;
		//console.log(thisdb.varname + " hashtableentries: " + hashtableentries);

		///////////////
		// HANDLE SEARCH FOR * SPECIALLY
		if (given === '*') {
			for (allpageix = 0; allpageix < thisdb.pages.length - 1; allpageix++) {	// Ignore last empty page
				results.hits.push([allpageix + 1, thisdb, 0]);
				//console.log(" ALL - " + JSON.stringify(allpageix + 1));
			}
			results.expression = [];
			var sw = new SearchWord('*', '*', OP_NONE, false);
			sw.anyfound = true;
			results.expression.push(sw);
			return;
		}

		///////////////
		// PARSE INPUT (ONCE ONLY)
		////// BuildSearchWords
		// ' \x27  \x2018  \x2019	quote
		// " \x22  \x201C  \x201D	quote
		if (!results.expression) {
			expr = [];
			word = "";
			var op = OP_NONE;
			incontig = false;
			for (chno = 0; chno < given.length; chno++) {
				nch = given.charCodeAt(chno);
				//log("nch=" + nch + " - " + isUnicodeSeparator(nch)+isUnicodeLetter(nch)+isUnicodeNumber(nch));
				if (isUnicodeSeparator(nch)) {
					if (word.length === 0) {
						if (thisdb.haswordposns && (nch === 0x22 || nch === 0x201c)) {
							incontig = true;
						}
					}
					else if (word.length > 0) {
						useRules = true;
						var endcontig = false;
						if (thisdb.haswordposns && incontig && (nch === 0x22 || nch === 0x201d)) {
							endcontig = true;
						}
						// Check for words in single quotes, ie look for ' before start of current word or  for 
						if (nch === 0x27) {
							if ((chno > word.length) && (given.charCodeAt(chno - word.length - 1) === 0x27)) {
								useRules = false;
							}
						}
						if (nch === 0x2019) {
							if ((chno > word.length) && (given.charCodeAt(chno - word.length - 1) === 0x2018)) {
								useRules = false;
							}
						}
						//log("word=" + word + " - " + useRules);
						op = storeWord(results.narrative, expr, word, op, incontig, useRules);
						word = "";
						if (endcontig) { incontig = false; }
					}
				}
				else {
					if (!isUnicodeCharLatin(nch)) {
						if (word.length > 0) {
							op = storeWord(results.narrative, expr, word, op, incontig, useRules);
							word = "";
						}
					}
					word += String.fromCharCode(nch);
				}
			}
			if (word.length > 0) {
				//log("word=" + word);
				storeWord(results.narrative, expr, word, op, incontig);
			}
			results.expression = expr;
		}

		///////////////
		// FIND PAGE LISTS
		//log("FIND PAGE LISTS");
		// For each given search word
		for (eno = 0; eno < results.expression.length; eno++) {
			searchWord = results.expression[eno];
			searchWord.pl = [];
			var w = deepArrayCopy(searchWord.wds);
			searchWord.wordfound = false;
			//log(searchWord.given);

			// For each variant of each given word
			for (wno = 0; wno < w.length; wno++) {
				word = w[wno];

				lcword = toUnicodeLowerCase(word);

				// look through hashtable for word
				hashcode = stringHashCode(lcword) % hashtableentries;
				do {
					wordList = thisdb.words[hashcode];
					//console.log(wordList);
					if (Object.prototype.toString.call(wordList) !== "[object Array]") {
						w.splice(wno, 1);
						wno--;
						break; // word not found
					}
					if (lcword === wordList[0]) {
						// word found
						// Copy and remove word from 1st element ie from ["quoted",[3,[[40,813]]]]
						prioList = wordList.slice(1);
						searchWord.pl.push(prioList);
						searchWord.wordfound = true;
						searchWord.anyfound = true;
						var inwdsfound = false;
						for (i = 0; i < searchWord.wdsfound.length; i++) {
							if (searchWord.wdsfound[i] === word) {
								inwdsfound = true;
							}
						}
						if (!inwdsfound) { searchWord.wdsfound.push(word); }
						break;
					}
					// No match, try next hash
					if (++hashcode === hashtableentries) {
						hashcode = 0;
					}
				}
				while (true);
			}
		}

		//log(expression);

		///////////////
		////// DoSearch

		// MERGE WORD PRIORITIES
		//log("MERGE WORD PRIORITIES");
		// For each given search word (into first word of each expression element)
		for (eno = 0; eno < results.expression.length; eno++) {
			searchWord = results.expression[eno];
			if (!searchWord.wordfound) {
				if (config.dbcount > 1) {
					return;	// Don't add any hits to results
				}
				continue;
			}
			searchWord.pr0 = deepArrayCopy(searchWord.pl[0]);
			if (searchWord.pl.length > 1) {
				// Merge subsequent alt-words into first alt-word
				//var plnono = 1;
				for (plno = 1; plno < searchWord.pl.length; plno++) {
					pln = searchWord.pl[plno];
					for (prno = 0; prno < pln.length; prno++) {
						prlist = pln[prno];
						prio = prlist[0];
						foundprio = false;
						for (pr0no = 0; pr0no < searchWord.pr0.length; pr0no++) {
							if (searchWord.pr0[pr0no][0] === prio) {
								// merge pagelists and ensure unique
								var concatted = searchWord.pr0[pr0no][1].concat(deepArrayCopy(prlist[1]));
								searchWord.pr0[pr0no][1] = plNormalise(concatted);
								foundprio = true;
								break;
							}
						}
						if (!foundprio) {
							searchWord.pr0.push(deepArrayCopy(prlist));
						}
					}
					// Remove this pl[plno]
					searchWord.pl.splice(plno, 1);
					plno--;
				}
				// Sort pageno lists
				for (pr0no = 0; pr0no < searchWord.pr0.length; pr0no++) {
					pr0list = searchWord.pr0[pr0no];
					pr0list[1].sort(function (a, b) { return a[0] - b[0]; });
				}
				// Sort priority lists
				searchWord.pr0.sort(function (a, b) { return a[0] - b[0]; }); // Sort by priority
			}

			////
			// Produce combined page list for this searchWord including best priority for each page
			//console.log("searchWord.pr0.length: " + searchWord.pr0.length);
			// combined has array of	[pageno,prio,wordposn1...]
			//										or	[pageno,prio,totalwordcount]
			searchWord.combined = [];
			for (pr0no = 0; pr0no < searchWord.pr0.length; pr0no++) {
				prio = searchWord.pr0[pr0no][0];
				pl = searchWord.pr0[pr0no][1];
				for (i = 0; i < pl.length; i++) {
					pl[i].splice(1, 0, prio);
				}
				// Then add onto end of combined
				searchWord.combined = searchWord.combined.concat(pl);
			}
			searchWord.combined = plNormaliseAndSortPositions(searchWord.combined);
			searchWord.combined.sort(function (a, b) { return a[0] - b[0]; });

		}

		///////////////
		////// MERGE PAGES
		//	Handle:	logical operations ie OR NOT
		//					contiguous words
		hitinfo = [];	// array of [pageno,bestprio,totalwordcount]
		anyhits = false;
		incontig = 0;
		var contigs = false;
		for (eno = 0; eno < results.expression.length; eno++) {
			searchWord = results.expression[eno];
			if (!searchWord.wordfound) { continue; }	// Carry on if word not found at all

			//console.log("MERGE: " + searchWord.given + " op:" + searchWord.op);

			if (searchWord.op & OP_CONTIG) {
				incontig++;
				if (incontig === 1) {
					contigs = searchWord.combined;	//	array of [pageno,prio,wordposn1...]
				}
			} else {
				incontig = 0;
				contigs = false;
			}

			if (searchWord.op & OP_NOT) {
				var notted = [];
				pageix = 0;
				pl = searchWord.combined[pageix];
				for (pageno = 1; pageno < thisdb.pages.length; pageno++) {	// Note start and end
					if (pl[0] == pageno) {
						if (++pageix < searchWord.combined.length) {
							pl = searchWord.combined[pageix];
						}
					}
					else {
						notted.push([pageno, 3, 0]);
					}
				}
				searchWord.combined = notted;
			}
			if (!anyhits) {
				for (pageix = 0; pageix < searchWord.combined.length; pageix++) {
					pl = searchWord.combined[pageix];
					totalwordcount = (thisdb.haswordposns ? pl.length - 2 : pl[2]);
					hitinfo.push([pl[0], pl[1], totalwordcount]);
				}
				anyhits = true;
				continue;
			}
			// go through hitinfo and tomerge to create merged, then set hitinfo to merged
			var isOr = (searchWord.op & OP_OR);
			tomerge = searchWord.combined;
			pageix = 0;
			mergeix = 0;
			merged = [];
			while ((pageix < hitinfo.length) && (mergeix < tomerge.length)) {
				pageno = hitinfo[pageix][0];
				mergepageno = tomerge[mergeix][0];
				if (pageno < mergepageno) {
					if (isOr) {
						merged.push([pageno, hitinfo[pageix][1], hitinfo[pageix][2]]);
					}
					pageix++;
					continue;
				}
				if (pageno > mergepageno) {
					if (isOr) {
						pl = tomerge[mergeix];
						totalwordcount = (thisdb.haswordposns ? pl.length - 2 : pl[2]);
						merged.push([mergepageno, tomerge[mergeix][1], totalwordcount]);
					}
					mergeix++;
					continue;
				}
				// pageno == mergepageno
				bestprio = Math.min(hitinfo[pageix][1], tomerge[mergeix][1]);

				var updatedContigs, contigfound;
				var totalwordcount2 = 0;
				if (incontig > 1) {
					var c1i = findConfig(pageno, contigs);														// [6,3,111,159,190,291,375]
					var c1 = contigs[c1i];
					totalwordcount2 = -(c1.length - 2);
					var c2i = findConfig(mergepageno, searchWord.combined);					// [6,0,17,70,112,265,292,372,388,459,477,520,555,592,671]
					var c2 = searchWord.combined[c2i];
					updatedContigs = [pageno, bestprio];												// [6,0,112,292]
					contigfound = false;
					var c2pno = 2;
					for (var c1pno = 2; c1pno < c1.length; c1pno++) {
						var c1p = c1[c1pno];
						while (c2pno < c2.length) {
							var c2p = c2[c2pno];
							if (c2p < c1p) {
								c2pno++;
							} else if (c2p == (c1p + 1)) {
								contigfound = true;
								updatedContigs.push(c2p);
								break;
							} else {
								break;
							}
						}
					}
					if (!contigfound) {
						pageix++;
						mergeix++;
						continue;
					} else {
						contigs[c1i] = updatedContigs;
						totalwordcount2 += updatedContigs.length - 2;
					}
				}

				if (!contigfound) {
					totalwordcount2 = (thisdb.haswordposns ? tomerge[mergeix].length - 2 : tomerge[mergeix][2]);
				}
				totalwordcount = hitinfo[pageix][2] + totalwordcount2;
				merged.push([pageno, bestprio, totalwordcount]);
				pageix++;
				mergeix++;
			}
			// If ORing then add in the ones we've missed
			if (isOr) {
				for (; pageix < hitinfo.length; pageix++) {
					merged.push([hitinfo[pageix][0], hitinfo[pageix][1], hitinfo[pageix][2]]);
				}
				for (; mergeix < tomerge.length; mergeix++) {
					pl = tomerge[mergeix];
					totalwordcount = (thisdb.haswordposns ? pl.length - 2 : pl[2]);
					merged.push([tomerge[mergeix][0], tomerge[mergeix][1], totalwordcount]);
				}
			}
			hitinfo = merged;
		}

		///////////////
		// Build results hits to return
		for (var hitno = 0; hitno < hitinfo.length; hitno++) {
			pageno = hitinfo[hitno][0];
			bestprio = hitinfo[hitno][1];
			totalwordcount = hitinfo[hitno][2];
			var relevance = (3 - bestprio) * 10000 + totalwordcount;
			results.hits.push([pageno, thisdb, relevance]);
		}
	}

	/////////////////////////////////////
	function findConfig(pageno, contigs) {
		for (var i = 0; i < contigs.length; i++) {
			if (contigs[i][0] === pageno) {
				return i;
			}
		}
		throw "findConfig " + pageno + " fail";
	}

	/////////////////////////////////////
	isUnicodeSeparator = function (nch) {
		return !isUnicodeLetter(nch) && !isUnicodeNumber(nch);
	};

	/////////////////////////////////////
	isUnicodeLetter = function (nch) {
		var ChTopByte = (nch >> 8);

		if (ChTopByte === 0) {
			if ((nch >= 0x61 && nch <= 0x7a) || // a z
			(nch >= 0x41 && nch <= 0x5a) || // A Z
			(nch === 0x00AA) ||
			(nch === 0x00B5) ||
			(nch === 0x00BA)) {
				return true;
			}
			if (nch >= 0x00C0 && (nch !== 0x00D7 && nch !== 0x00F7)) {
				return true;
			}
			return false;
		}
		if ((nch >= 0x4E00 && nch <= 0x9FA5) || // CJK Ideographs
		(nch >= 0xAC00 && nch <= 0xD7A3) ||			// Hangul syllables
		(nch >= 0xF900 && nch <= 0xFA2D)) {			// CJK compatibility ideographs
			return true;
		}
		// ??? More tests needed using LetterRange
		return false;
	};

	/////////////////////////////////////
	isUnicodeNumber = function (nch) {
		//log("nch=" + nch);
		var ChTopByte = (nch >> 8);

		if (ChTopByte === 0) {
			if ((nch >= 0x30 && nch <= 0x39) || // 0 9
			(nch >= 0x00B2 && nch <= 0x00B3) ||
			(nch === 0x00B9) ||
			(nch >= 0x00BC && nch <= 0x00BE)) {
				return true;
			}
			return false;
		}
		// ??? More tests needed using NumberRange
		return false;
	};

	/////////////////////////////////////
	isUnicodeCharLatin = function (nch) {
		return (nch <= 0x02A8) ||
				(nch >= 0x1E00 && nch < 0x1EF9) ||
				(nch >= 0x0374 && nch <= 0x03f5) ||
				(nch >= 0x0400 && nch <= 0x04ff) ||
				(nch >= 0x060C && nch <= 0x06FE) ||
				(nch >= 0xFB50 && nch <= 0xFDFB) ||
				(nch >= 0xFE70 && nch <= 0xFEFC);
	};

	/////////////////////////////////////
	toUnicodeLowerCase = function (word) {
		// ???
		return word.toLowerCase();
	};

	/////////////////////////////////////
	// Amended from http://stackoverflow.com/questions/7616461/generate-a-hash-from-string-in-javascript-jquery

	stringHashCode = function (val) {
		var hash = 0, i, chr, len;
		for (i = 0, len = val.length; i < len; i++) {
			chr = val.charCodeAt(i);
			hash = ((hash << 5) - hash) + chr;
			hash &= 0x7fffffff; // Convert to 32bit integer
		}
		return hash;
	};

	/////////////////////////////////////
	// PUBLIC methods
	return {
		version: version,
		versionDate: versionDate,
		search: search,
		config: config,
		databases: databases,
		OP_AND: OP_AND,
		OP_NONE: OP_NONE,
		OP_OR: OP_OR,
		OP_NOT: OP_NOT,
		OP_CONTIG: OP_CONTIG
	};
	/////////////////////////////////////
}(jQuery);